/** @file camogm_align.c
 * @brief Provides frame alignment functions use for recording to block device.
 * @copyright Copyright (C) 2017 Elphel, Inc.
 *
 * @par <b>License</b>
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/** @brief This define is needed to use lseek64 and should be set before includes */
#define _LARGEFILE64_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/uio.h>

#include "camogm_align.h"

static inline void vectcpy(struct iovec *dest, void *src, size_t len);
static inline void vectset(struct iovec *dest, int src, size_t len);
static inline void vectshrink(struct iovec *vec, size_t len);
static inline void vectshrinkhead(struct iovec *vec, size_t len);
static inline unsigned char *vectrpos(struct iovec *vec, size_t offset);
/* debug functions */
static void dev_dbg(const char *prefix, const char *format, ...);


/** Replace debug function with the same name from driver code with debug macro to reduce changes */
static void dev_dbg(const char *prefix, const char *format, ...)
{
	va_list args;

	va_start(args, format);
	D6(vfprintf(debug_file, format, args));
	va_end(args);
}

/** Copy @e len bytes from buffer pointed by @e src vector to buffer pointed by @e dest vector */
static inline void vectcpy(struct iovec *dest, void *src, size_t len)
{
	unsigned char *d = (unsigned char *)dest->iov_base;
	if (len <= 0) return;
	memcpy(d + dest->iov_len, src, len);
	dest->iov_len += len;
}

/** Set @e len bytes with @e src to buffer pointed by @e dest vector */
static void vectset(struct iovec *dest,  int src, size_t len)
{
	unsigned char *d = (unsigned char *)dest->iov_base;
	D6(fprintf(debug_file, "_13c06a?_:vectset() d=%p, src=%d len=%d\n",	d, src, len ));

	if (len <= 0) return;
	memset(d + dest->iov_len, src, len);
	dest->iov_len += len;
}



/** Shrink vector length by @len bytes */
static inline void vectshrink(struct iovec *vec, size_t len)
{
	if (vec->iov_len >= len) {
		vec->iov_len -= len;
	}
}

/** Shrink vector length by @len bytes from the head */
static inline void vectshrinkhead(struct iovec *vec, size_t len)
{
	if (vec->iov_len >= len) {
		vec->iov_len -= len;
		unsigned char * p = (unsigned char * ) vec->iov_base; // += len;
		vec->iov_base = p + len;
	}
}


/** This helper function is used to position a pointer @e offset bytes from the end
 * of a buffer. */
static inline unsigned char *vectrpos(struct iovec *vec, size_t offset)
{
	return (unsigned char *)vec->iov_base + (vec->iov_len - offset);
}

/** extend to page-aligned */
int vectaligntail(struct iovec *dest)
{
//	unsigned char *d = (unsigned char *)dest->iov_base;
	int len = dest->iov_len;
	int ceil_size = (len & (PAGE_PHYS - 1)) ? ((len | (PAGE_PHYS - 1)) + 1) : len;
	if (ceil_size > len) {
		vectset(dest,  0, ceil_size - len);
	}
	return ceil_size;
}



/** Remap vectors pointing to various buffers with frame data to vectors used during frame alignment
 *  Setup state->writer_params.data_segments[seg_page] ([SEGMENT_GLUE, SEGMENT_FIRST, SEGMENT_SECOND] and
 *  use /setup next state->writer_params.glue_carry_vec
 *  [SEGMENT_GLUE] may heve zero lengtgh - check all lengths before writev()
 *  */
size_t remap_vectors(camogm_state *state) // , struct iovec *chunks)
{
	// state->writer_params.glue_carry_vec has data for the the trailer of the previous frame. swap it with the
	// state->writer_params.glue_buffs[seg_page][SEGMENT_GLUE]
	size_t total_sz;
	struct writer_params *params = &state->writer_params;
	int seg_page = params->chunk_page_prep;
	long head_bytes, tail_bytes, trimmed_bytes;
	int  i, data_index, data_pre, num_glue_pages, trim_segment, circbuff_offs;
	int gap_size;
	unsigned char * aligned_data;
	struct iovec  glue_vec;
	struct iovec  * glue_vec_p; //  = &glue_vec;
	glue_vec.iov_base = params->glue_carry_vec.iov_base;
	glue_vec.iov_len = params->glue_carry_vec.iov_len;
	D7(fprintf(debug_file, "_13c01_:remap_vectors @ %07d\n",get_fpga_usec(state->fd_fparmsall[0], 0)));
	params->glue_carry_vec.iov_base = params->data_segments[seg_page][SEGMENT_GLUE].iov_base;
	params->glue_carry_vec.iov_len = 0; // params->data_segments[seg_page][SEGMENT_GLUE].iov_len;
	params->data_segments[seg_page][SEGMENT_GLUE].iov_base =  glue_vec.iov_base;
	params->data_segments[seg_page][SEGMENT_GLUE].iov_len =   glue_vec.iov_len;
	params->data_segments[seg_page][SEGMENT_FIRST].iov_len =  0; // check iov_len first before using iov_base
	params->data_segments[seg_page][SEGMENT_SECOND].iov_len = 0; // check iov_len first before using iov_base
	// calculate total header size
	head_bytes = 0;
	for (i = 0; i < state->chunk_data_index; i++){
		head_bytes += state->packetchunks[i].bytes;
	}
	D7(fprintf(debug_file, "_13c02_:remap_vectors, head_bytes=%ld @ %07d\n", head_bytes,get_fpga_usec(state->fd_fparmsall[0], 0)));

	// find out how much should be cut from the beginning of the first data segment (circbuf itself is page-aligned)
	// in an unlikely case that image is smaller that a page ...
	// state->port_num
	data_index = state->chunk_data_index;
	circbuff_offs = (int) state->packetchunks[data_index].chunk;
	aligned_data = state->packetchunks[data_index].chunk+ (PAGE_PHYS_CEIL(circbuff_offs) - circbuff_offs);
	data_pre = aligned_data - state->packetchunks[data_index].chunk;
	glue_vec_p = &params->data_segments[seg_page][SEGMENT_GLUE];
	D7(fprintf(debug_file, "_13c03_:remap_vectors, circbuff_offs=0x%x aligned_data=%p data_pre=0x%x @ %07d\n", circbuff_offs, aligned_data, data_pre, get_fpga_usec(state->fd_fparmsall[0], 0)));
	D7(fprintf(debug_file, "_13c04_:remap_vectors, state->packetchunks[state->chunk_data_index].bytes=%ld @ %07d\n", state->packetchunks[state->chunk_data_index].bytes, get_fpga_usec(state->fd_fparmsall[0], 0)));
	if (state->packetchunks[state->chunk_data_index].bytes < data_pre) {
		// Too short frame data - special treatment with zero data, moving some to carry, and possibly outputing some from the GLUE
		// there is definitely no second data - otherwise the first segment would end at page boundary
		// copy everything to params->data_segments[seg_page][SEGMENT_GLUE], trim to pages, pass residual to
		// params->glue_carry_vec
		// copy header data, no gap
		for (i = 0; i <= state->chunk_index; i++){ // including data and trailer
			vectcpy(glue_vec_p, state->packetchunks[i].chunk, state->packetchunks[i].bytes);
		}
		trimmed_bytes = PAGE_PHYS_TRIM(glue_vec_p->iov_len);
		D7(fprintf(debug_file, "_13c05_:remap_vectors, trimmed_bytes=%ld @ %07d\n", trimmed_bytes, get_fpga_usec(state->fd_fparmsall[0], 0)));
		if ((glue_vec_p->iov_len - trimmed_bytes) > 0){
			vectcpy(&params->glue_carry_vec,
					((char *) glue_vec_p->iov_base) + trimmed_bytes,
					glue_vec_p->iov_len - trimmed_bytes);
			glue_vec_p->iov_len = trimmed_bytes; // may be 0
		}
	} else {
		head_bytes += data_pre;
		tail_bytes = glue_vec_p->iov_len;
		num_glue_pages = PAGE_PHYS_NUMBER(tail_bytes + head_bytes);
		gap_size = num_glue_pages * PAGE_PHYS -  tail_bytes - head_bytes;
		D7(fprintf(debug_file, "_13c06_:remap_vectors, head_bytes=%ld, tail_bytes=%ld, num_glue_pages=%d, gap_size=%d @ %07d\n", head_bytes, tail_bytes, num_glue_pages, gap_size, get_fpga_usec(state->fd_fparmsall[0], 0)));
		// fill gap with zeros
		D7(fprintf(debug_file, "_13c06a_:remap_vectors, glue_vec_p=%p, gap_size=%d glue_vec_p->iov_base=0x%p, glue_vec_p->iov_len=%d @ %07d\n",\
				glue_vec_p, gap_size, glue_vec_p->iov_base, glue_vec_p->iov_len, get_fpga_usec(state->fd_fparmsall[0], 0)));
		D7(fprintf(debug_file, "_13c06a0_:malloc_usable_size(glue_vec_p->iov_base) = %d @ %07d\n", malloc_usable_size(glue_vec_p->iov_base), get_fpga_usec(state->fd_fparmsall[0], 0)));
		for (i = 0; i <  SEGMENTS_PAGES; i++){
			D7(fprintf(debug_file, "_13c06a1_: malloc_usable_size([%d][SEGMENT_GLUE]) = %d @ %07d\n",\
					i, malloc_usable_size(params->data_segments[i][SEGMENT_GLUE].iov_base), get_fpga_usec(state->fd_fparmsall[0], 0)));
		}

		// size_t malloc_usable_size(void *ptr);
		vectset(glue_vec_p,  0, gap_size);
		D7(fprintf(debug_file, "_13c06a2_:remap_vectors, glue_vec_p=%p, gap_size=%d glue_vec_p->iov_base=0x%p, glue_vec_p->iov_len=%d @ %07d\n",\
				glue_vec_p, gap_size, glue_vec_p->iov_base, glue_vec_p->iov_len, get_fpga_usec(state->fd_fparmsall[0], 0)));
		// copy header data after the gap
		for (i = 0; i < state->chunk_data_index; i++){
			D7(fprintf(debug_file, "_13c06b_:remap_vectors, i=%d glue_vec_p=%p, gap_size=%d glue_vec_p->iov_base=0x%p, glue_vec_p->iov_len=%d state->packetchunks[i].chunk=%p, state->packetchunks[i].bytes=%ld @ %07d\n", \
					i, glue_vec_p, gap_size, glue_vec_p->iov_base, glue_vec_p->iov_len, \
					state->packetchunks[i].chunk, state->packetchunks[i].bytes, \
					get_fpga_usec(state->fd_fparmsall[0], 0)));
			vectcpy(glue_vec_p, state->packetchunks[i].chunk, state->packetchunks[i].bytes);
		}
		// copy (unaligned) beginning of the data segment
		vectcpy(glue_vec_p, state->packetchunks[state->chunk_data_index].chunk, data_pre);
		// TODO: verify that glue_vec_p->iov_len is multiple of PAGE_PHYS
		trim_segment = SEGMENT_FIRST;
		params->data_segments[seg_page][SEGMENT_FIRST].iov_base = ((char *) state->packetchunks[state->chunk_data_index].chunk) + data_pre;
		params->data_segments[seg_page][SEGMENT_FIRST].iov_len =            state->packetchunks[state->chunk_data_index].bytes - data_pre;
		D7(fprintf(debug_file, "_13c07_:remap_vectors @ %07d\n", get_fpga_usec(state->fd_fparmsall[0], 0)));
		if (params->data_segments[seg_page][SEGMENT_FIRST].iov_len == 0) {
			// nothing left of data segment 1 - move segment2 to segment1 (or skip both)
			if (state->data_segments > 1) { // move second segment to first
				params->data_segments[seg_page][SEGMENT_FIRST].iov_base = ((char *) state->packetchunks[state->chunk_data_index + 1].chunk);
				params->data_segments[seg_page][SEGMENT_FIRST].iov_len =            state->packetchunks[state->chunk_data_index + 1].bytes ;
			}
		} else if (state->data_segments > 1) {
			params->data_segments[seg_page][SEGMENT_SECOND].iov_base = ((char *) state->packetchunks[state->chunk_data_index + 1].chunk);
			params->data_segments[seg_page][SEGMENT_SECOND].iov_len =            state->packetchunks[state->chunk_data_index + 1].bytes ;
			trim_segment = SEGMENT_SECOND;
		}
		D7(fprintf(debug_file, "_13c08_:remap_vectors,[SEGMENT_GLUE].iov_len=%d, [SEGMENT_FIRST].iov_len=%d, [SEGMENT_SECOND].iov_len=%d  @ %07d\n", \
				params->data_segments[seg_page][SEGMENT_GLUE].iov_len, \
				params->data_segments[seg_page][SEGMENT_FIRST].iov_len, \
				params->data_segments[seg_page][SEGMENT_SECOND].iov_len, \
				get_fpga_usec(state->fd_fparmsall[0], 0)));

		// Trim last data segment to PAGE_PHYS (it may make its length zero - OK), copy remaining
		// data to the buffer pointed by params->glue_carry_vec.iov_base, set params->glue_carry_vec.iov_len
		trimmed_bytes = PAGE_PHYS_TRIM(params->data_segments[seg_page][trim_segment].iov_len);
		D7(fprintf(debug_file, "_13c09_:remap_vectors, trimmed_bytes=%ld @ %07d\n", trimmed_bytes, get_fpga_usec(state->fd_fparmsall[0], 0)));
		// debugging dma-related bugs.
		params->dbg_data[seg_page][0] = state->port_num;
		params->dbg_data[seg_page][1] = params->data_segments[seg_page][trim_segment].iov_len - trimmed_bytes; // remainder copied to glue buffer
		if (params->data_segments[seg_page][trim_segment].iov_len > trimmed_bytes) {
			D6(fprintf(debug_file, "_13c09a:remap_vectors, trim_segment=%d, dest = %p, dest_len=%d src=%p, copy_len = %ld @ %07d\n", \
					trim_segment,
					&params->glue_carry_vec, \
					params->glue_carry_vec.iov_len, \
					((char *) params->data_segments[seg_page][trim_segment].iov_base) + trimmed_bytes, \
					params->data_segments[seg_page][trim_segment].iov_len - trimmed_bytes, \
					get_fpga_usec(state->fd_fparmsall[0], 0)));

			vectcpy(&params->glue_carry_vec,
					((char *) params->data_segments[seg_page][trim_segment].iov_base) + trimmed_bytes,
					params->data_segments[seg_page][trim_segment].iov_len - trimmed_bytes);
			params->data_segments[seg_page][trim_segment].iov_len = trimmed_bytes; // reduce data segment length
		}


		D7(fprintf(debug_file, "_13c09b_:remap_vectors, tail_segm=%d, chunk_index=%d @ %07d\n", \
				state->chunk_data_index + state->data_segments, \
				state->chunk_index, \
				get_fpga_usec(state->fd_fparmsall[0], 0)));
		// FIXME: next can only be tested with JPEG/JP4 !
		for (i = state->chunk_data_index + state->data_segments; i < state->chunk_index; i++){ // trailer if it exists
			D7(fprintf(debug_file, "_13c09c_:remap_vectors, i=%d, dest = %p, dest_len=%d src=%p, len =%ld @ %07d\n", \
					i, \
					&params->glue_carry_vec, \
					params->glue_carry_vec.iov_len, \
					state->packetchunks[i].chunk, \
					state->packetchunks[i].bytes, \
					get_fpga_usec(state->fd_fparmsall[0], 0)));

			vectcpy(&params->glue_carry_vec,
					state->packetchunks[i].chunk,
					state->packetchunks[i].bytes);
		}
	}
	D7(fprintf(debug_file, "_13c10_:remap_vectors @ %07d\n", get_fpga_usec(state->fd_fparmsall[0], 0)));
	total_sz = 0; // get_size_from_paged(state, 0, INCLUDE_REM) + rbuff->iov_len;
	for (i = 0; i < SEGMENTS_NUMBER; i++){
		total_sz += params->data_segments[seg_page][i].iov_len;
	}
	//port_num
	state->image_size[state->port_num] = total_sz;
	D4(fprintf(debug_file, "total_sz = %d\n", total_sz));
	return total_sz;
}

/** Allocate and initialize buffers for frame alignment */
int init_align_buffers(camogm_state *state)
{
	struct writer_params *params = &state->writer_params;
	int seg_page;
    D3(fprintf(debug_file, "GLUE_SEGMENTS = %d pages of %d bytes each\n", GLUE_SEGMENTS, PAGE_PHYS));
	for (seg_page = 0; seg_page < SEGMENTS_PAGES + 1; seg_page++){
		// allocate SEGMENT_GLUE buffer that will contain data copied from non-pagfe aligned segments
		params->glue_buffs[seg_page] = (unsigned char *) aligned_alloc (PAGE_PHYS, GLUE_SEGMENTS * PAGE_PHYS);
		if (params->glue_buffs[seg_page] == NULL) {
			return -1;
		}
		D6(fprintf(debug_file, "init_align_buffers(), params->glue_buffs[%d]=%p @ %07d\n", \
				seg_page, params->glue_buffs[seg_page], get_fpga_usec(state->fd_fparmsall[0], 0)));
	}
	for (seg_page = 0; seg_page < SEGMENTS_PAGES; seg_page++){
		params->data_segments[seg_page] = (struct iovec *)malloc(SEGMENTS_NUMBER * sizeof(struct iovec));
		if (params->data_segments[seg_page] == NULL) {
			return -1;
		}
		D6(fprintf(debug_file, "init_align_buffers(), seg_page=%d, params->data_segments[seg_page]=%p @ %07d\n", \
				seg_page, params->data_segments[seg_page], get_fpga_usec(state->fd_fparmsall[0], 0)));

		params->data_segments[seg_page][SEGMENT_GLUE].iov_base = params->glue_buffs[seg_page];
		params->data_segments[seg_page][SEGMENT_GLUE].iov_len =  0;
	}
	params->glue_carry_vec.iov_base = params->glue_buffs[SEGMENTS_PAGES];
	params->glue_carry_vec.iov_len = 0; // nothing there fro the first frame (TODO: need same for start write?)

	return 0;
}

/** Delete buffers for frame alignment */
void deinit_align_buffers(camogm_state *state)
{
	int seg_page;
	struct writer_params *params = &state->writer_params;
	for (seg_page = 0; seg_page < SEGMENTS_PAGES+1; seg_page++){
		free (params->glue_buffs[seg_page]);
		params->glue_buffs[seg_page] = NULL;
	}
	for (seg_page = 0; seg_page < SEGMENTS_PAGES; seg_page++){
		free(params->data_segments[seg_page]);
		params->data_segments[seg_page] = NULL;
	}
}

void reset_segments(camogm_state *state, int all, int page)
{
	int i;
	struct writer_params *params = &state->writer_params;
	struct iovec *vects = params->data_segments[page];
	for (i = 0; i < SEGMENTS_NUMBER; i++) {
		vects[i].iov_len = 0;
	}
	if (all) {
		params->glue_carry_vec.iov_len = 0;
	}
}

/** Convert LBA to byte offset used for lseek */
uint64_t lba_to_offset(uint64_t lba)
//uint64_t lba_to_offset(uint64_t lba)
{
	return lba * PHY_BLOCK_SIZE;
}

/** Get next LBA after segment prepared for writing. There is no actual
 * write pointer as recording can be out of order. When the "file" is closed,
 * this pointer matches the normal file pointer, just in blocks.
 * @param params record state parameters
 * @return full LBA (from the disk start) write pointer, measured in 512-byte sectors.
 */
uint64_t get_lba_next(const struct writer_params *params){
	return params->lba_start + (params->next_segment_pos / PHY_BLOCK_SIZE); // next prepared (not yet written)
}

